from typing import Dict

import logging
import time

from django.conf import settings
from mailersend import emails
from celery import shared_task


logger = logging.getLogger('emails')


@shared_task(name='djmailer_send_email')
def send_email_by_mailersend(mail_body: Dict) -> None:
    sending = True
    email = [to_mail['email'] for to_mail in mail_body['to']]
    mailer = emails.NewEmail(settings.MAILERSEND_API_KEY)
    while sending:
        code, result = mailer.send(mail_body).split('\n')
        if int(code) == 202:
            logger.info(f'Sent an email to {email}')
            sending = False
        elif int(code) == 429:
            logger.debug('Waiting 5 seconds before new try...')
            time.sleep(5)
        else:
            logger.error(
                f'Something went wrong: {email}, '
                f'result: {result}, code: {code}'
            )
            sending = False



from unittest.mock import patch

from django.core.mail.message import EmailMultiAlternatives
from django.test.utils import override_settings
from django.core.mail import send_mail
from django.test import TestCase

from djmailer.backend import EmailBackend
from djmailer.tasks import send_email_by_mailersend


@override_settings(
    DEBUG=True,
    EMAIL_BACKEND='djmailer.backend.EmailBackend',
    DJMAILER_FROM_EMAIL='verified@mail.com',
    DJMAILER_FROM_NAME='Verified Name',
)
class TestMailerEmailBackend(TestCase):
    path_to_celery_task = 'djmailer.tasks.send_email_by_mailersend.delay'
    path_to_pr_send_email = 'djmailer.backend.EmailBackend._send_email'

    @override_settings(DEBUG=True)
    def test_pr_send_email_if_debug_true(self) -> None:
        with patch(self.path_to_celery_task) as send_email_task:
            send_mail(
                'Subject here',
                'Here is the message.',
                'from@example.com',
                ['to1@example.com', 'to2@example.com'],
            )
            send_email_task.assert_not_called()

    @override_settings(DEBUG=False)
    def test_pr_send_email_if_debug_false(self) -> None:
        with patch(self.path_to_celery_task) as send_email_task:
            send_mail(
                'Subject here',
                'Here is the message.',
                'from@example.com',
                ['to1@example.com', 'to2@example.com'],
            )
            send_email_task.assert_called_once()

    def test_message_sent_by_backend(self) -> None:
        with patch('djmailer.backend.EmailBackend.send_messages') as email_send:
            send_mail(
                'Subject here',
                'Here is the message.',
                'from@example.com',
                ['to1@example.com', 'to2@example.com'],
            )
            email_send.assert_called_once()

    def test_email_with_html(self) -> None:
        with patch(self.path_to_pr_send_email) as send_email:
            send_mail(
                'Subject here',
                'Here is the message.',
                'from@example.com',
                ['to1@example.com', 'to2@example.com'],
                html_message='<h1>Hello!</h1>'
            )
            send_email.assert_called_once()
            args, _ = send_email.call_args
            mail_body = args[0]
            self.assertIn('html', mail_body.keys())
            self.assertEqual(mail_body['html'], '<h1>Hello!</h1>')

    def test_bytes_attachments(self) -> None:
        with patch(self.path_to_pr_send_email) as send_email:
            mail = EmailMultiAlternatives(
                'Subject here',
                'Here is the message.',
                to=['to1@example.com', 'to2@example.com'],
            )
            mail.attach('hello.txt', b'Hello world!', 'image/png')
            mail.send()
            args, _ = send_email.call_args
            mail_body = args[0]
            self.assertIn('attachments', mail_body.keys())
            attached_file = mail_body['attachments'][0]
            self.assertIn('content', attached_file.keys())
            self.assertEqual('SGVsbG8gd29ybGQh', attached_file['content'])
            send_email.assert_called_once()

    def test_plain_attachment(self) -> None:
        with patch(self.path_to_pr_send_email) as send_email:
            mail = EmailMultiAlternatives(
                'Subject here',
                'Here is the message.',
                to=['to1@example.com', 'to2@example.com'],
            )
            mail.attach('hello.txt', 'Hello world!', 'text/plain')
            mail.send()
            args, _ = send_email.call_args
            mail_body = args[0]
            self.assertIn('attachments', mail_body.keys())
            attached_file = mail_body['attachments'][0]
            self.assertIn('content', attached_file.keys())
            self.assertEqual('SGVsbG8gd29ybGQh', attached_file['content'])
            send_email.assert_called_once()

    def test_multiple_mailing(self) -> None:
        with patch(self.path_to_pr_send_email) as send_email:
            mail = EmailMultiAlternatives(
                'Subject here',
                'Here is the message.',
                to=[f'to{i}@example.com' for i in range(200)],
            )
            mail.attach('hello.txt', b'Hello world!', 'image/png')
            mail.send()
            self.assertEqual(send_email.call_count, 4)

    def test_empty_to_email_mailing(self) -> None:
        with patch(self.path_to_pr_send_email) as send_email:
            mail = EmailMultiAlternatives(
                'Subject here',
                'Here is the message.',
                to=[],
            )
            mail.attach('hello.txt', b'Hello world!', 'image/png')
            mail.send()
            send_email.assert_not_called()


class TestCeleryTasks(TestCase):
    mail = EmailMultiAlternatives(
            'Subject here',
            'Here is the message.',
            'from@example.com',
            ['to@email.com', ],
        )
    mail_body = EmailBackend()._parse_email_message(mail)

    def test_send_email_by_mailersend_202(self) -> None:
        with self.assertLogs('emails', level='DEBUG') as email_logger:
            with patch('mailersend.emails.NewEmail.send') as mock:
                mock.return_value = '202\nAccepted'
                send_email_by_mailersend(self.mail_body)
                mock.assert_called_once_with(self.mail_body)
                self.assertEqual(
                    email_logger.output,
                    [f'INFO:emails:Sent an email to {["to@email.com", ]}']
                )

    def test_send_email_by_mailersend_429(self) -> None:
        with patch('time.sleep') as sleep:
            with self.assertLogs('emails', level='DEBUG') as email_logger:
                with patch('mailersend.emails.NewEmail.send') as mock:
                    mock.side_effect = [
                        '429\nToo Many Requests',
                        '202\nAccepted'
                    ]
                    send_email_by_mailersend(self.mail_body)
                    mock.assert_called_with(self.mail_body)
                    self.assertEqual(mock.call_count, 2)
                    sleep.assert_called_once_with(5)
                    self.assertEqual(
                        email_logger.output, [
                            'DEBUG:emails:Waiting 5 seconds before new try...',
                            f'INFO:emails:Sent an email to {["to@email.com", ]}'
                        ]
                    )

    def test_send_email_by_mailersend_error(self) -> None:
        with self.assertLogs('emails', level='DEBUG') as email_logger:
            with patch('mailersend.emails.NewEmail.send') as mock:
                mock.return_value = '400\nBad Request'
                send_email_by_mailersend(self.mail_body)
                mock.assert_called_with(self.mail_body)
                self.assertEqual(
                    email_logger.output, [
                        f'ERROR:emails:Something went wrong: '
                        f'{["to@email.com", ]}, '
                        'result: Bad Request, code: 400'
                    ]
                )



from typing import Dict, List, Union

from base64 import b64encode
import logging

from django.core.mail.backends.base import BaseEmailBackend
from django.core.mail.message import EmailMultiAlternatives
from django.conf import settings

from djmailer.tasks import send_email_by_mailersend


logger = logging.getLogger('emails')


class EmailBackend(BaseEmailBackend):
    @staticmethod
    def _send_email(mail_body: Dict) -> None:
        to_emails = [to_mail['email'] for to_mail in mail_body['to']]
        mail_subject = mail_body['subject']
        if not settings.DEBUG:
            logger.info(f"Sending email to {to_emails}"
                        f" with subject {mail_subject}...")
            send_email_by_mailersend.delay(mail_body)
            logger.info(f"Task has been assigned to"
                        f"celery: email to {to_emails}")
        else:
            logger.debug(f"Sending email to {to_emails}"
                         f" with subject {mail_subject}")

    @staticmethod
    def _convert_email_address_for_mailersend(
                email: Union[str, List[str]]
            ) -> List[Dict[str, str]]:
        return [{'email': email_address, } for email_address in email]

    def _parse_email_message(
                self, django_email: EmailMultiAlternatives
            ) -> Dict:
        mail_body = {
            'text': django_email.body,
            'subject': django_email.subject,
            'from': {
                    'email': settings.DJMAILER_FROM_EMAIL,
                    'name': settings.DJMAILER_FROM_NAME,
                },
            'to': self._convert_email_address_for_mailersend(
                    django_email.to
                ),
            'reply_to': self._convert_email_address_for_mailersend(
                    django_email.reply_to
                ),
            'cc': self._convert_email_address_for_mailersend(
                    django_email.cc
                ),
            'bcc': self._convert_email_address_for_mailersend(
                    django_email.bcc
                ),
        }

        for content, mimetype in django_email.alternatives:
            if mimetype == 'text/html':
                mail_body['html'] = content
        if django_email.attachments:
            mail_body['attachments'] = list()
            for filename, content, _ in django_email.attachments:
                if isinstance(content, str):
                    content = content.encode('utf-8')
                mail_body['attachments'].append({
                    'filename': filename,
                    'content': b64encode(content).decode('utf-8'),
                    'disposition': 'attachment'
                })
        return mail_body

    def _send(self, email_message: EmailMultiAlternatives) -> None:
        recipients = email_message.recipients()
        # if there are more than 50 recipients
        # we need to divide them by 50
        # since mailersend has a limit
        if len(recipients) > 50:
            chunked_emails = [recipients[i:i + 50]
                              for i in range(0, len(recipients), 50)]
            for email in chunked_emails:
                email_message.to = email
                mail_body = self._parse_email_message(email_message)
                self._send_email(mail_body)
        else:
            mail_body = self._parse_email_message(email_message)
            self._send_email(mail_body)

    def send_messages(
                self, email_message: List[EmailMultiAlternatives]
            ) -> None:
        if len(email_message) == 0:
            return  # pragma: no cover
        for message in email_message:
            self._send(message)
